
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#define _GNU_SOURCE
#include <stdio.h>
#include <alloca.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <assert.h>
#include <fnmatch.h>

#include "initng_process_db.h"
#include "initng_service_cache.h"
#include "initng_global.h"
#include "initng_string_tools.h"
#include "initng_active_db.h"
#include "initng_toolbox.h"
#include "initng_plugin_callers.h"

#include "initng_service_data_types.h"
#include "initng_struct_data.h"

/*
 * This is a simple list_add(&service->list), but
 * will first check for dupletts.
 */
int service_db_add(service_h * s)
{
    service_h *current = NULL;

    assert(s);

    /* Check that we are unique */
    if ((current = service_db_find_by_name(s->name)))
    {
        D_("service_db_add(%s): WARNING: DUPLICATE. Already added!\n",
           s->name);
        return (FALSE);
    }

    /* if not found, add this one on the list */
    list_add(&(s->list), &(g.service_cache.list));

    return (TRUE);
}


/*
 * Instead of service_db_new, this service allocates a new
 * service, and copy data from another service.
 *
 * OBS, this is not, as setting service->father.
 */
service_h *service_db_copy(char *name, service_h * s)
{
    service_h *new_serv = NULL;

    assert(s);
    assert(name);
    D_("copy_service(%s);\n", s->name);

    /* allocate new service */
    new_serv = (service_h *) initng_calloc(1, sizeof(service_h));
    if (!new_serv)
    {
        F_("Cound not allocate space, for new copy of service!\n");
        return (NULL);
    }

    /* copy service structure */
    memcpy(new_serv, s, sizeof(service_h));
    /* name has to diff */
    new_serv->name = name;

    /* initziate the data structure */
    INIT_LIST_HEAD(&(new_serv->data.list));

    /* copy over all strings */
    d_copy_all(&s->list, &new_serv->list);

    return (new_serv);
}

/*
 * Allocates an entry, in memory, and sets defaults,
 * name & type, initziate lits.
 */
service_h *service_db_new(char *name, e_type type)
{
    service_h *new_s = NULL;

    assert(name);
    D_("default_service();\n");
    if (!(new_s = (service_h *) initng_calloc(1, sizeof(service_h))))
    {
        F_("OUT OF MEMORY default_service().\n");
        return (NULL);
    }

    /* set name & type */
    new_s->name = name;
    new_s->type = type;

    /* initziate the data list struct */
    INIT_LIST_HEAD(&(new_s->data.list));
    return (new_s);
}

/*
 * Search the service-db, look for an exact match
 * on service name, returns a pointer to it.
 */
service_h *service_db_find_by_exact_name(char *service)
{
    service_h *current = NULL;

    assert(service);

    while_service_db(current)
    {
        /* first, try to find exact match */
        if (strcmp(service, current->name) == 0)
        {
            D_("service_db_find_by_name(%s): Found exact match\n",
               current->name);
            return (current);
        }
    }

    return (NULL);
}

service_h *service_db_find_by_name(char *service)
{

    assert(service);

    service_h *current = service_db_find_by_exact_name(service);

    if (current)
        return (current);

    /* do pattern matching only if service name does not contain wildcards */
    if (strchr(service, '*') || strchr(service, '?'))
        return (NULL);

    current = NULL;

    /* create service name mask for pattern matching */
    char *service_mask = initng_calloc(strlen(service) + 3, sizeof(char));

    strcpy(service_mask, "*/");
    strcat(service_mask, service);

    while_service_db(current)
    {
        /* then try to find alike name */

        /* if existing service is a pattern (i.e. net ) */
        if (strchr(current->name, '*') || strchr(current->name, '?'))
        {
            if (fnmatch(current->name, service, FNM_CASEFOLD) == 0)

            {
                D_("service_db_find_by_name(%s): Found\n", current->name);
                free(service_mask);
                return (current);
            }
        }
        else if (fnmatch(service_mask, current->name, FNM_CASEFOLD) == 0)
        {
            D_("service_db_find_by_name(%s): Found\n", current->name);
            free(service_mask);
            return (current);
        }
    }

    free(service_mask);
    return (NULL);
}

/*
 * Search the service_db, first looking for an exact
 * match, and if not found, serches for string in
 * name, and if found it returns pointer.
 */
service_h *service_db_find_in_name(char *find)
{
    service_h *current = NULL;

    assert(find);

    /* First search exakt match */
    while_service_db(current)
    {
        if (fnmatch(current->name, find, FNM_CASEFOLD) == 0)
        {
            D_("service_db_find_by_name(%s): Found\n", current->name);
            return (current);
        }
    }

    current = NULL;
    while_service_db(current)
    {
        /* then try to find alike name ( service/net/ * ) */
        if (fnmatch(current->name, find, FNM_CASEFOLD) == 0)
        {
            D_("service_db_find_by_name(%s): Found\n", current->name);
            return (current);
        }
    }

    current = NULL;
    while_service_db(current)
    {
        if (strstr(current->name, find))
        {
            D_("service_db_find_by_name(%s): Found\n", current->name);
            return (current);
        }
    }

    return (NULL);
}


/*
 * When a service is freed, we have to make sure there is
 * no service that links to empty space, so we walk
 * the service_db, looking for this service pointer, 
 * and removes it.
 */
static void clear_father_pointer_from(service_h * s)
{
    service_h *current = NULL;

    while_service_db(current)
    {
        if (current->father == s)
            current->father = NULL;
    }
}

/*
 * Free all dynamically allocated strings, and other data
 * in this service, and free the service.
 * OBS. this function WONT delete the entry from the
 * service cache, so calling this, with this service on
 * that list, will make initng segfault.
 */
int service_db_free(service_h * to_free)
{
    assert(to_free);
    D_("service_db_free(%s);\n", to_free->name);

    /* change other database relations to this service to null */
    active_db_change_service_h(to_free, NULL);

    /* this is a dangerous pointer that dont have to hang loose */
    clear_father_pointer_from(to_free);

    /* free all data & strings in service */
    d_remove_all(&to_free->data.list);
    INIT_LIST_HEAD(&(to_free->data.list));

    /* free service name */
    if (to_free->name)
        free(to_free->name);
    free(to_free);
    return (TRUE);
}

/*
 * "Free all service data" == "Flush the service data cache"
 */
void service_db_free_all(void)
{
    service_h *current, *q = NULL;

    while_service_db_safe(current, q)
    {
        list_del(&current->list);
        service_db_free(current);
    }
}

/*
 * if the service->father_name is set, but the pointer
 * to the service service->father is not set, this
 * functions go out, searching for that father.
 */
int service_db_find_father(service_h * service)
{
    assert(service);
    /* if fahter is known, return ok */
    if (service->father)
        return (TRUE);
    /* if no father is needed */
    if (!service->father_name)
        return (FALSE);

    /* whats for father name in service db */
    service->father = service_db_find_by_name(service->father_name);
    if (service->father)
        return (TRUE);

    /* try to load father */
    service->father = initng_modules_parse(service->father_name);
    if (service->father)
        return (TRUE);

    /* else fail */
    return (FALSE);
}










/*
 *  DATA WRAPPERS BELOW:
 *
 * These are wrappers, that cant be made macros in initng_service_cache.h
 * wraps to initng_struct_data.c types.
 *
 *
 */

/*
 * service_db_get_string(), look in the service_h->data entry, but also
 * checks fathers->data, that will also apply for this service.
 */
char *service_db_get_string(s_entry * type, service_h * s)
{
    char *tmp = NULL;

    assert(type);
    assert(s);

    /* try to find in local data */
    if ((tmp = d_get_string(type, &s->data.list)))
        return (tmp);

    /* look for data in my parrent */
    if (service_db_find_father(s) && s->father)
        if ((tmp = service_db_get_string(type, s->father)))
            return (tmp);

    return (NULL);
}


/*
 * service_db_get_int(), look in the service_h->data entry, but also
 * checks fathers->data, that will also apply for this service.
 */
int service_db_get_int(s_entry * type, service_h * s)
{
    assert(type);
    assert(s);
    /* try to find in local data */
    if (d_is(type, &s->data.list))
        return (d_get_int(type, &s->data.list));
    /* look for data in my parrent, this is bad and wont look in parents parent. */
    if (service_db_find_father(s) && s->father
        && d_is(type, &s->father->data.list))
        return (d_get_int(type, &s->data.list));
    return (0);
}

/*
 * service_db_get_next_string(), look in the service_h->data entry, but also
 * checks fathers->data, that will also apply for this service.
 */
char *service_db_get_next_string(s_entry * type, service_h * s, char *last)
{
    char *tmp = NULL;

    assert(type);
    assert(s);
    /* try to find in local data */
    if ((tmp = d_get_next_string(type, &s->data.list, last)))
        return (tmp);
    /* look for data in my parrent */
    if (service_db_find_father(s) && s->father)
        if ((tmp = service_db_get_next_string(type, s->father, last)))
            return (tmp);
    return (NULL);
}

/*
 * service_db_is(), look in the service_h->data entry, but also
 * checks fathers->data, that will also apply for this service.
 */
int service_db_is(s_entry * type, service_h * s)
{
    assert(type);
    assert(s);
    if (d_is(type, &s->data.list))
        return (TRUE);
    /* look for data in my parrent */
    if (service_db_find_father(s) && s->father)
        if (service_db_is(type, s->father))
            return (TRUE);
    return (FALSE);
}

